/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include "version.h"
#include "reader.h"
#include "error.h"
#include "vfile.h"
#include "stack.h"
#include "refer.h"
#include "image.h"
#include "include.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <set>
#include <map>
#include <deque>


////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class related_section_t
{
public:
  std::string fname;
  std::string indexname;
  std::string aname;
  std::string sectionname;
  bool   refer;
};

static std::multimap<std::string,related_section_t> related_sections;
static std::set<std::string> related_section_keys;

static void related_section_regist(std::string key, std::string fname, std::string indexname , std::string aname, std::string sectionname,bool refer)
{
  related_section_t rel;

  rel.fname = fname;
  rel.indexname = indexname;
  rel.aname   = aname;
  rel.sectionname = sectionname;
  rel.refer = refer;

  related_section_keys.insert(key);
  related_sections.insert( std::pair<std::string,related_section_t>(key,rel) );

}


static void related_sections_manipulate()
{
  for( std::set<std::string>::iterator key = related_section_keys.begin(); key!=related_section_keys.end(); key++){
    
    int n = related_sections.count( *key );
    //printf("%d\n",n);
    if( n < 2 ){
      printf("WARNING: relation %s is appeared but only once.\n",(*key).c_str());
      continue;
    }

    related_section_t **rels = new related_section_t* [n];
    std::multimap<std::string,related_section_t>::iterator p = related_sections.find(*key);
    
    int i=0;
    do{
      rels[i++] = &(p->second);
      p++;
    }while( p != related_sections.upper_bound(*key) );

    for(i=0;i<n;i++){
      related_section_t *r1=rels[i];
      VFile *vf = vf_search(r1->fname.c_str());
      vf->seek(r1->indexname);

      int n_relation = 0;
      for(int j=0;j<n;j++){
	if( i==j || rels[j]->refer )
	  continue;
	n_relation++;
      }

      if( n_relation > 0 ){ 
	vf->printf("%s",TEXT_RELATION);
	vf->puts("<table><tr><td>");
	vf->puts("<ul>");
	for(int j=0;j<n;j++){
	  if( i==j )
	    continue;
	  related_section_t *r2 = rels[j];
	  if( r2->refer )
	    continue;
	  vf->printf("<li><a href=\"%s#%s\">%s</a><br>\n" ,
		     r2->fname.c_str(),
		     r2->aname.c_str(),
		     r2->sectionname.c_str()
		     );
	}
	vf->puts("</ul>\n");
	vf->puts("</td></tr></table>\n");
      }

    }
    
    delete rels;
  }
}

/////////////////////////////////////////////////////////////////////////////////////////



static VFile *indexhtml;


static void htheader(VFile *out,const char *title)
{

  out->printf("<html lang=\"%s\">\n",LANG);
  out->printf("<!-- Generated by ezht %s.\nezht COPYRIGHT (C) Tradcrafts.\n",VERSION);
  out->puts("-->\n");

  out->puts("<head>\n");
  out->printf("<meta name=\"generator\" content=\"ezht version %s\">\n",VERSION);
  out->printf("<meta name=\"content-language\" content=\"%s\">\n",LANG);
  out->printf("<meta http-equiv=\"content-type\" content=\"text/html; charset=%s\">\n",CHARSET);
  out->printf("<title lang=\"%s\">%s</title>\n",LANG,title);
  out->puts(HT_HEAD);
  out->puts("</head>\n");
  out->printf("<body %s>\n",BODY_ATTRIBUTES);
  out->puts(BODY_HEADER);
}

static void htfooter(VFile *out)
{
  out->puts(BODY_FOOTER);
  out->puts("</body>\n");
  out->puts("</html>\n");
}


static bool check(const char *line,VFile*,const char*,bool,std::string parent,std::string aname,std::string secname,std::set<VFile*> *localindexes,
		  std::list<std::string> *autopop_list);

static std::deque<int> numque;
static std::string mk_secnumtext(const char *number_separator)
{
  std::string secnumtext;
  std::deque<int>::iterator it=numque.begin();

  while(1){
    char tmp[16]; sprintf(tmp,"%d",(int)*it);
    secnumtext += tmp;
    it++;
    if( it != numque.end() )
      secnumtext += number_separator;
    else
      break;
  }

  return secnumtext;
}

static const char *get_newfilename()
{
  std::string num_s;
  
  const char *rule = FILENAME_RULE;
  if( 0==strcmp("serial",rule) ){
    char sbuf[8];
    static int serial_begin = 1;
    sprintf(sbuf,"%05d",serial_begin++);
    num_s = sbuf;
  }else if( 0==strcmp("section",rule) ){
    num_s = mk_secnumtext("-");
  }else{ // "line"
    num_s = reader.number_info();
  }

  static std::string buf;
  buf = FILENAME_PREFIX + num_s + FILENAME_SUFFIX;

  return buf.c_str();
}


static const char *get_newaname()
{
  static char sbuf[16];
  sprintf(sbuf,"sec%s",reader.number_info().c_str());
  return sbuf;
}



static std::string prev_section_href;
static std::set<VFile*> prev_next_vfiles;

static const char *mk_tmpindex()
{
  static int id=0;
  static char tmp[32];
  sprintf(tmp,"tmpidx.%d",id++);
  return tmp;
}

static void write_head_of_section(VFile *vf,std::string aname,std::string name,std::string secnumtxt,std::string parent)
{
  vf->printf("<a name=\"%s\"></a>\n",aname.c_str());

  vf->puts(prev_section_href.c_str());

  VFile *cvf = vf->clone();
  const char *tmpidx = mk_tmpindex();
  cvf->add_index(tmpidx); cvf->seek(tmpidx);
  prev_next_vfiles.insert(cvf);

  vf->printf("<a href=\"%s\" style=\"text-decoration: none\" title=\"Up\">%s</a> ", parent.c_str() ,ANCHOR_UP );
  vf->printf("<a href=\"JavaScript:history.go(-1)\" style=\"text-decoration: none\" title=\"Back\">%s</a> ",ANCHOR_BACK);
  vf->printf("<a href=\"%s#%s\" style=\"text-decoration: none\">%s</a>",INDEXFILE,aname.c_str(),ANCHOR_INDEX);
  vf->printf("<br>");

  vf->printf("<div style=\"background-color: %s; padding: 2px 2px 2px 2px;\">%s%s <strong>%s</strong></div>\n",
	     SECTION_BAR_COLOR,SECTION_MARK,secnumtxt.c_str(),name.c_str());

  vf->puts("<div align=right>");

  vf->puts(prev_section_href.c_str());

  cvf = vf->clone();
  tmpidx = mk_tmpindex();
  cvf->add_index(tmpidx); cvf->seek(tmpidx);
  prev_next_vfiles.insert(cvf);

  vf->printf("<a href=\"%s\" style=\"text-decoration: none\" title=\"Up\">%s</a> ", parent.c_str(), ANCHOR_UP);
  vf->printf("<a href=\"JavaScript:history.go(-1)\" style=\"text-decoration: none\" title=\"Back\">%s</a> ",ANCHOR_BACK);
  vf->printf("<a href=\"%s#%s\" style=\"text-decoration: none\">%s</a>",INDEXFILE,aname.c_str(),ANCHOR_INDEX);

  vf->puts("</div>");
  

}



static std::set<VFile*> indexfiles;
static void section(const char *name,VFile *o,const char *filename,bool packed,bool inherit,std::string parent)
{
  std::list<std::string> autopop_list;

  bool opened = false;
  VFile *out;
  

  numque.back() ++;
  std::string secnumtxt = mk_secnumtext(NUMBER_SEPARATOR);

  std::string aname = get_newaname();

  std::string fnam;
  if( ! o || ! packed || ! inherit ){
    fnam = get_newfilename();
    out = vf_open(fnam.c_str());
    opened = true;
    std::string title = (std::string)DOCNAME + " - " + name;
    htheader(out,title.c_str());
  }else{
    out = o; 
    fnam = filename!=NULL? filename : "" ; 
    out->puts("\n");
  }




  
  std::string href;
  if( packed )
    href = fnam + '#' + aname;
  else
    href = fnam;

  for(std::set<VFile*>::iterator itr=prev_next_vfiles.begin(); itr!=prev_next_vfiles.end(); itr++){
    VFile *pnvf = *itr;
    pnvf->printf("<a href=\"%s\" style=\"text-decoration: none\" title=\"Next - %s\">%s</a> ",href.c_str(),name,ANCHOR_NEXT);
    delete pnvf;
  }
  prev_next_vfiles.clear();


  write_head_of_section(out,aname,name,secnumtxt,parent);
  
  bool o_in_indexfiles = false;
  if( o!=NULL && o != out ){
    write_head_of_section(o,aname,name,secnumtxt,parent);
    o->printf("<p>%s <a href=\"%s\">&quot;<b>%s</b>&quot;</a></p>\n",TEXT_SECTION_DISJOINED,href.c_str(),name);
    o->puts("<ol>\n");
    indexfiles.insert(o);
    o_in_indexfiles = true;
  }

  prev_section_href = "<a href=\"" + href + "\" style=\"text-decoration: none\" title=\"Back - " + name + "\">" + ANCHOR_PREV + "</a> ";
  

  for(std::set<VFile*>::iterator indexfp=indexfiles.begin(); indexfp!=indexfiles.end(); indexfp++){
    (*indexfp)->printf("<li><a name=\"%s\" href=\"%s\">%s</a>\n",
		       (*indexfp==indexhtml?aname.c_str():"") ,
		       href.c_str(),
		       name);
      (*indexfp)->puts("<ol>\n");
  }

  std::set<VFile*> localindexes;
  numque.push_back(0);
  bool success = false;
  const char *line;
  while( NULL != (line=getline(true)) ){
    if( 0==strcmp(line,"@end") ){
      success = true;
      break;
    }
    else if( ! check(line,out,fnam.c_str(),packed,fnam+'#'+aname,aname,name,&localindexes,&autopop_list) ){
      out->puts(line);
    }
  }
  numque.pop_back();


  if( opened ){
    htfooter(out);
  }

  for(std::set<VFile*>::iterator indexfp=indexfiles.begin(); indexfp!=indexfiles.end(); indexfp++){
    (*indexfp)->puts("</ol>\n");
  }

  if( o_in_indexfiles ){
    o->puts("</ol>\n");
    indexfiles.erase(o);
  }

  for(std::set<VFile*>::iterator idx=localindexes.begin();idx!=localindexes.end();idx++){
    indexfiles.erase( *idx );
    (*idx)->puts("</td></tr></table>\n");
    delete *idx;
  }


  for(std::list<std::string>::iterator itr=autopop_list.begin(); itr!=autopop_list.end(); itr++)
    stack_pop((*itr).c_str());


  if( ! success )
    warn("WARNING: unexpected end of file.\n");
}




static bool check(const char *line,VFile *o,const char *fnam,bool packed,std::string parent,
	   std::string aname,std::string sectionname,std::set<VFile*> *localindexes,std::list<std::string> *autopop_list)
{
  if( line[0]=='@' ){
    std::vector<std::string> vec;
    parse(line,vec);
    
    std::string cmd = vec[0];
    
    if( cmd == "push" ){
      cmd_push(vec);
      return true;
    }
    else if( cmd == "pop" ){
      cmd_pop(vec);
      return true;
    }else if( cmd == "section" ){
      if( vec.size() > 3 ){ error("@section: too many parameters"); return true; }
      if( vec.size() == 1){ error("@section: parameter required"); return true; }

      if( vec.size() == 2 ){
	section(vec[1].c_str(),o,fnam,packed,true,parent);
      }else if( vec[2] == "joined" ){
	section(vec[1].c_str(),o,fnam,true,false,parent);
      }else if( vec[2] == "disjoined" ){
	section(vec[1].c_str(),o,fnam,false,false,parent);
      }else{
	error("@section: unknown joint `%s'",vec[2].c_str());
	return true;
      }

      return true;
    }else if( cmd == "relate" ){
      if( !o ){ error("misplaced @relate"); return true; }
      if( vec.size() > 3 ){ error("@relate: too many parameters"); return true; }
      if( vec.size() == 1){ error("@relate: parameter required"); return true; }
      bool refer = false;
      if( vec.size()==3 ){
	if( vec[2] == "refer" )
	  refer = true;
	else{
	  error("@relate: unknown option `%s'",vec[2].c_str()); return true;
	}
      }
      static int uniq_id=0;
      char idx_name[16];
      sprintf(idx_name,"relat.%d",uniq_id++);
      o->add_index(idx_name);
      related_section_regist(vec[1], fnam, idx_name , aname, sectionname, refer);
      return true;
    }else if( cmd == "index" ){
      if( !o ){ error("misplaced @index"); return true; }
      if( vec.size() > 1 ){ error("@index: unknown parameter specified"); return true; }
      VFile *tmpvf = o->clone();
      static int uniq_id=0;
      char idx[16];
      sprintf(idx,"index.%d",uniq_id++);
      tmpvf->add_index(idx);
      tmpvf->seek(idx);

      indexfiles.insert(tmpvf);
      localindexes->insert(tmpvf);
      tmpvf->puts("<table><tr><td>\n");
      tmpvf->puts("<ol>\n");
      return true;
    }else if( cmd == "autopop" ){
      if( !autopop_list ){ error("misplaced @autopop"); return true; }
      cmd_autopop(*autopop_list,vec);
      return true;
    }else if( cmd == "mark" ){
      cmd_mark(vec,o,fnam,aname);
      return true;
    }
    else{
      error("@%s: Unknown command",cmd.c_str()); 
    }
  }
  
  
  return false;

}


void generate()
{

  indexhtml = vf_open(INDEXFILE); // Ǥϥե̾ϻŪʤΤǤ(ǥեȤindex.html)

  indexfiles.insert(indexhtml);

  
  indexhtml->add_index("head");
  indexhtml->puts("<ol>\n");

  numque.push_back(0);
  const char *buf;
  while( NULL != (buf=getline(false)) ){
    if( ! check(buf,NULL,NULL,false,INDEXFILE,"","",NULL,NULL) ){
      error("Command expected");
    }
  }

  indexhtml->puts("</ol>\n");
  htfooter(indexhtml);
  
  indexhtml->seek("head");
  std::string title = (std::string)DOCNAME + " - " + TEXT_INDEX; 
  htheader(indexhtml,title.c_str());
  indexhtml->puts(INDEXFILE_INTRODUCTION);

  indexhtml->reset_filename(INDEXFILE); // ե̾ꤵ줿Τѹ(ѹƤʤindex.htmlΤޤ)

  related_sections_manipulate();
  fixate_references();
  fixate_images();

  fixate_jumps();

}





