/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include "error.h"
#include "image.h"
#include "vfile.h"

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <stdint.h>

static inline void inv(void *p,int size)
{
  int i=0,j=size-1,m=size/2;
  while(i<m)
    {
      int8_t tmp = *(((int8_t*)p)+i);
      *(((int8_t*)p)+i++) = *(((int8_t*)p)+j);
      *(((int8_t*)p)+j--) = tmp;
    }
}

static int16_t lendian16(int16_t num)
{
  uint16_t check = 0xff00;
  if( *(uint8_t*)&check ){
    // BIG ENDIANξ
    inv(&num,sizeof(num));
  }

  return num;
}

static int32_t bendian32(int32_t num)
{
  uint16_t check = 0x00ff;
  if( *(uint8_t*)&check ){
    // LITTLE ENDIANξ
    inv(&num,sizeof(num));
  }

  return num;
}

static int16_t bendian16(int16_t num)
{
  uint16_t check = 0x00ff;
  if( *(uint8_t*)&check ){
    // LITTLE ENDIANξ
    inv(&num,sizeof(num));
  }

  return num;
}

static bool get_image_info(const char *fname, int *width_r, int *height_r)
{
  FILE *f = fopen(fname,"rb");
  if( ! f ){
    warn("@image: can't open `%s': %s",fname,strerror(errno));
    return false;
  }

  bool success = false;

  char head[4];
  fread(head,4,1,f);
  
  if( 0 == memcmp("GIF",head,3) ){
    // GIF
    fseek(f,6,SEEK_SET);
    int16_t width,height;
    fread(&width,2,1,f); 
    fread(&height,2,1,f); 
    *width_r = lendian16(width);
    *height_r= lendian16(height);
    success=true;
  }else if( 0 == memcmp("PNG",head+1,3) ){
    // PNG
    fseek(f,8,SEEK_SET);
    int32_t off;
    char    code[4];
    while( fread(&off,4,1,f)==1 && fread(code,4,1,f)==1 ){
      off = bendian32(off);
      if( 0 == memcmp(code,"IHDR",4) ){
	int32_t width,height;
	fread(&width,4,1,f); 
	fread(&height,4,1,f); 
	*width_r = bendian32(width);
	*height_r= bendian32(height);
	success = true;
	break;
      }else if( 0 == memcmp(code,"IEND",4) ){
	break;
      }
      else{
	fseek(f,off+4,SEEK_CUR);
      }
    }
  }else{
    fseek(f,6,SEEK_SET);
    char tmp[4];
    fread(tmp,4,1,f);
    if( 0 == memcmp(tmp,"JFIF",4) ){
      // JPEG
      int b;
      while( EOF != (b=fgetc(f)) ){
	if( b==255 ){
	  int t = fgetc(f);
	  int16_t size;
	  fread(&size,2,1,f);
	  size = bendian16(size);
	  if( t==0xc0 || t==0xc2){
	    int16_t width,height;
	    fgetc(f); // ΤƤ
	    fread(&height,2,1,f); 
	    fread(&width,2,1,f); 
	    *width_r = bendian16(width);
	    *height_r= bendian16(height);
	    success = true;
	    break;
	  }else{
	    fseek(f,size-2,SEEK_CUR);
	  }
	}
      }
      
    }
  }
  
  if( success && *width_r>0 && *height_r>0 )
    return true;

  warn("%s: could not take a image size",fname);
  return false;


}

#define IMGMARK "<!--img-->"

void cmd_image(std::vector<std::string> &args,std::string &ret_r)
{
  ret_r = ""; // ǥХå
  if( args.size() == 1 ){
    error("@image: no image specified"); return;
  }

  if( args.size() > 3 ){
    error("@image: too many arguments"); return;
  }

  std::string imgfile = args[1];
  std::string opt;
  if( args.size() == 3 )
    opt = args[2];
  
  int width,height;
  if( get_image_info(args[1].c_str(),&width,&height) ){
    char tmp[64];
    sprintf(tmp,"width=\"%d\" height=\"%d\" ",width,height);
    opt = tmp + opt;
  }

  ret_r = IMGMARK  "<img src=\"" + imgfile + "\" " + opt + '>';
}


static bool _fix_image_callback(VFile *vf,const char *line,std::string&,
				std::string*)
{
  if( strstr(line,IMGMARK) )
    vf->chomp(); // ľιԤβԤȤ

  return false;
}

void fixate_images()
{
  vf_each(_fix_image_callback);
}


