/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include "error.h"
#include "reader.h"
#include "stack.h"
#include "vfile.h"
#include <map>

#define SPSYMBOL "%file%"

static std::string filename_insert(const char *cmdline,const char *fname)
{
  const char *inspnt = strstr(cmdline,SPSYMBOL);
  if( ! inspnt ){
    error("@insert: incorrect include-filter `%s': symbol %s not found",cmdline,SPSYMBOL);
    return "";
  }

  char *tmp = new char[strlen(cmdline)+strlen(fname)]; // ɬ׽ʬʸĹȤʤ
  memcpy(tmp,cmdline,inspnt-cmdline);
  tmp[inspnt-cmdline] = '\0';
  strcat(tmp,fname);
  strcat(tmp,inspnt+strlen(SPSYMBOL));
  std::string ret = tmp;
  delete tmp;
  
  return ret;
}

void cmd_include(std::vector<std::string> &args)
{

  if( args.size() == 1 ){
    error("@include: no include file specified"); return;
  }

  if( args.size() > 3 ){
    error("@include: too many arguments"); return;
  }

  int include_opt = 0;
  
  if( args.size() == 3 ) {
    if( args[2] == "html" )
      include_opt = READOPT_INCLUDE_HTML;
    else if( args[2] == "numbered" )
      include_opt = READOPT_INCLUDE_NUMBERED;
    else if( args[2] == "text" )
      include_opt = READOPT_INCLUDE_TEXT;
    else{
      error("@include: Unknown option `%s'",args[2].c_str()); return;
    }
  }

  std::string fname;
  std::string filter = INCLUDE_FILTER;
  std::string filter_cmd;
  bool use_filter=false;

  
  fname = args[1];
  
  if( filter != "" ){
    // ե륿ꤵƤ
    filter_cmd = filename_insert(filter.c_str(),fname.c_str());
    use_filter = true;
  }

  if( ! reader.open(fname.c_str(),include_opt,(use_filter?filter_cmd.c_str():NULL)) ){
    error("@include: Unable to open `%s' ",fname.c_str());
  }
  
  reader.set_current_header( INCLUDE_HEADER );
  reader.set_current_footer( INCLUDE_FOOTER );
  reader.set_textmode();
}


static std::map< std::string , std::map<int,std::string*> > line_indexes;
static std::map< std::string , std::string > fname_table;

static bool _make_line_indexes_callback(VFile *vf, const char *line, std::string &modified_line_r, std::string *ref)
{
  static bool now_indexing = false;
  static std::map<int,std::string*> *curmap = NULL;
  static int curline = 0;

  if( now_indexing ){
    if( 0 == strcmp(line,READER_INCLUDE_END) ){
      now_indexing = false;
      modified_line_r = "";
      return true;
    }else{
      (*curmap)[curline++] = ref;
      return false;
    }
  }
  
  // ELSE

  
  if( line == strstr(line,READER_INCLUDE_BEGIN) ){
    const char *fname = line + strlen(READER_INCLUDE_BEGIN);
    //fprintf(stdout,"INC=[%s]\n",fname);
    curmap = &line_indexes[fname];
    curmap->clear();
    curline = 1;
    now_indexing = true;
    
    fname_table[fname] = vf->get_filename();

    modified_line_r = "";
    return true;
  }else
    return false;
  
}

static void make_line_indexes()
{
  vf_each(_make_line_indexes_callback);
}




struct jumpstruct_t
{
  int linenum;
  std::string fname;
  std::string attrs;
  std::string anchor;
};

#define JUMPMARK "\xff\xff@JMP\xff\xff"

void cmd_jump(std::vector<std::string> &args, std::string &ret)
{
  if( args.size() < 4 ){
    error("@jump: too few argument(s)");
    return;
  }else if( args.size() > 5 ){
    error("@jump: too many arguments");
    return;
  }

  std::string fname = args[1];
  int  linenum = atoi(args[2].c_str());
  std::string anchor = args[3];

  std::string attrs;
  if( args.size() == 5 )
    attrs = args[4];

  jumpstruct_t *struc = new jumpstruct_t;
  struc->fname = fname;
  struc->linenum = linenum;
  struc->attrs = attrs;
  struc->anchor = anchor;

  char tmp[64];
  sprintf(tmp,"%s%p:",JUMPMARK,struc);
  ret = tmp;
  
}


static bool _fixate_jump_callback(VFile *vf, const char *line, std::string &modified_line_r, std::string*)
{
  bool found = false;
  const char *s = line;
  const char *sub;

  while( NULL != (sub=strstr(s,JUMPMARK)) ){
    found = true;
    s = sub + strlen(JUMPMARK);

    jumpstruct_t *struc;
    sscanf(s,"%p",&struc);

    while( *s != ':' ) // skip
      s++;
    s++;

    std::map< std::string , std::map<int,std::string*> >::iterator p;
    p = line_indexes.find(struc->fname);
    if( p == line_indexes.end() ){
      error("@jump: unknown file `%s'",struc->fname.c_str());
      return false;
    }
    
    std::map<int,std::string*>::iterator result = p->second.find(struc->linenum);
    if( result == p->second.end() ){
      error("@jump: `%s' illegal line number [%d]: out of range",struc->fname.c_str(),struc->linenum);
      return false;
    }
    
    std::string *line_linked = result->second; 
    
    static int id=0;
    char aname[16];
    sprintf(aname,"jmp%d",id++);
    // Ƭ˥֤դ뤳Ȥաޤ,ιԤιƬϥ֤Ʒ
    *line_linked = (std::string)"\t<a name=\"" + aname + "\"></a>" + (line_linked->c_str()+1);
    
    modified_line_r += "<a href=\"" + fname_table[struc->fname] + '#' + aname + "\" " + struc->attrs + '>' + struc->anchor + "</a>";
  }

  if( found ){
    modified_line_r += s; // Ĥ򤯤äĤ
    vf->chomp();
    return true;
  }else
    return false;
}

void fixate_jumps()
{
  make_line_indexes();
  vf_each(_fixate_jump_callback);
}
