/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>


#include <string>
#include <vector>
#include <list>

#ifndef unix
#include <windows.h>
#endif

#include "version.h"
#include "reader.h"
#include "error.h"
#include "stack.h"
#include "gen.h"
#include "vfile.h"


static void usage(const char *prognam)
{
  fprintf(stderr,"Error - Incorrect argument(s).\n");
  fprintf(stderr,"usage:\n");
  fprintf(stderr,"%s srcfile\n",prognam);
  fprintf(stderr,"%s srcfile [outfile]\n",prognam);
  
#ifndef unix
  MessageBox(NULL,"Incorrect argument(s)\nSee document for details.","Error",MB_OK);
#endif
}

int main(int argc,char **argv)
{
  
  fprintf(stderr,"ezht version %s - Copyright (C) 2004 Tradcrafts.\n",VERSION);
  
  if( argc < 2 || argc > 3){
    usage("ezht");
    exit(-1);
  }

  stack_init();

  const char *infile = argv[1];
  const char *outfile = NULL;
  if( argc > 2 )
    outfile = argv[2];

  if( ! reader.open( infile ) ){
    fprintf(stderr,"unable to read `%s': %s\n",infile,strerror(errno));
    exit(-1);
  }
 
  generate();
  reader_tempfile_delete();

  if( argc == 2 )
    vf_flushall();
  else if( argv[2] == (std::string)"STDOUT" )
    vf_flushall_to(stdout);
  else{
    FILE *out = fopen(argv[2],"w");
    if( ! out ){
      fprintf(stderr,"unable to write `%s': %s\n",argv[2],strerror(errno));
    }else{
      vf_flushall_to(out);
      fclose(out);
    }
  }

  return 0; // Exit normally
}
