/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string>
#include <vector>

#include "reader.h"
#include "include.h"
#include "begintext.h"
#include "error.h"
#include "stack.h"
#include "image.h"
#include "refer.h"
#include "ezht_glob.h"

Reader reader;

Reader::Reader()
{
  sourcemode_columns = 0;
  sourcemode_line_cnt = 0;
}

Reader::~Reader()
{
}

void Reader::set_current_readopt(int opt)
{
  infoque[0].readopt = opt;
}

void Reader::set_sourcemode_columns()
{
  int n_line=0;
  int c;
  // Կ׻
  while( EOF != (c=get()) ){
    if( c=='\n' )
      n_line++;
  }
  rewind(); // rewind

  sourcemode_columns = 0;
  while( n_line != 0 ){
    sourcemode_columns ++;
    n_line /= 10;
  }

  sourcemode_line_cnt = 1; // initialize
}

std::string Reader::sourcemode_linenumber()
{
  char tmp[16];
  sprintf(tmp,"%d",sourcemode_line_cnt++);
  
  int numlen = strlen(tmp);
  int pudding = sourcemode_columns - numlen;

  char buf[256];
  buf[0] = '\0';
  while( pudding-- )
    strcat(buf,"&nbsp;");
  strcat(buf,tmp);
  strcat(buf,":&nbsp;");

  return buf;
}

bool Reader::open(const char *fnam,int readopt,const char *filter)
{
  FILE *newfp;
  if( 0 != strcmp(fnam,"STDIN") ){
    if( filter )
      newfp = popen(filter,"r");
    else
      newfp = fopen(fnam,"r");
  }
  else
    newfp = stdin;
    
  if( ! newfp ){
    // ERROR
    return false;
  }
  
  infoque.push_front( fileinfo_t(fnam,readopt) );
  fileinfo_t &info = infoque[0];
  
  int c;
  while( EOF != (c=fgetc(newfp)) )
    info.buf.push_back((char)c);
  rewind(); // ݥ󥿤Ƭ˥å

  if( newfp != stdin )
    fclose(newfp);

  if( readopt&READOPT_INCLUDE_NUMBERED )
    set_sourcemode_columns();

  return true;
}

int Reader::get()
{
  if( infoque.size() > 0 ){
    fileinfo_t &info = infoque[0];
    if( info.buf_pnt == info.buf.end() )
      return EOF;
    else
      return *info.buf_pnt++;
  }
  
  return EOF;
}

void Reader::rewind()
{
  if( infoque.size() > 0 ){
    infoque[0].buf_pnt = infoque[0].buf.begin();
  }
}

const char* Reader::get_fnam()
{
  if( infoque.size() > 0 )
    return infoque[0].fnam.c_str();
  else
    return NULL;
}

int Reader::get_linenum()
{
  if( infoque.size() > 0)
    return infoque[0].linenum;
  else
    return -1;
}

std::string Reader::number_info()
{
  std::string num_info;

  int i = (int)infoque.size() - 1;
  while(i >= 0){
    char tmp[16];
    sprintf(tmp,"%d",infoque[0].linenum);
    if( num_info == "" )
      num_info = tmp;
    else
      num_info += (std::string)tmp + '-';
    i--;
  }
    
  return num_info;
}


static int _mb_check_euc(unsigned char c)
{
  if( c == 0x8F ) // (3Хʸ)
    return 3;
  else if( c >= 0xA1 && c <= 0xFE ) // 2Хʸ
    return 2;
  else
    return 1; // Ⱦ
}

static int _mb_check_sjis(unsigned char c)
{
  if( c >= 0x81 && c <= 0x9F ||
      c >= 0xE0 && c <= 0xEF )
    return 2; // Хʸ
  else
    return 1; // Ⱦ or Ⱦѥ
}

static int _mb_check_iso88591(unsigned char c)
{
  return 1;
}

static char* _ins_esc(char *s,const char *repstr)
{
  int repstr_len = strlen(repstr);
  memmove(s+repstr_len,s+1,strlen(s+1)+1);
  memcpy(s,repstr,repstr_len);
  return s+repstr_len;
}

static void _text_chars_escape(char *s)
{
  int (*mb_check)(unsigned char);

  const char *charset = CHARSET;
  if( strcasecmp(charset,"EUC-JP") == 0 )
    mb_check = _mb_check_euc;
  else if( strcasecmp(charset,"Shift_JIS") == 0 )
    mb_check = _mb_check_sjis;
  else if( strcasecmp(charset,"iso-8859-1") == 0 )
    mb_check = _mb_check_iso88591;
  else{
    error("unknown charset `%s'",charset); return;
  }


  std::string tabspace = "";
  
  unsigned char c;
  while( '\0' != (c=(unsigned char)*s) ){
    int chlen = (*mb_check)(c);
    if( chlen > 1 ){
      for(int i=0; i<chlen && *s; i++,s++)
	; // skip
    }else{ // 1Хʸ
      switch( (char)c ){
      case '<': s = _ins_esc(s,"&lt;"); break;
      case '>': s = _ins_esc(s,"&gt;"); break;
      case '"': s = _ins_esc(s,"&quot;"); break;
      case '&': s = _ins_esc(s,"&amp;"); break;
      case '\n': s= _ins_esc(s,"<br>\n"); break;
      default:
	if( isspace((char)c) ){
	  if( (char)c != '\t' )
	    s = _ins_esc(s,"&nbsp;");
	  else{
	    if( tabspace=="" ){
	      int tab_spaces = atoi(TAB_SPACES);
	      for(int i=0;i<tab_spaces;i++)
		tabspace += "&nbsp;";
	    }
	    s = _ins_esc(s,tabspace.c_str());
	  }
	}
	else if( !isascii((char)c) ){
	  char tmp[8];
	  sprintf(tmp,"&#%d;",(int)c);
	  s = _ins_esc(s,tmp);
	}
	else
	  s++;
	break;
      }
    }
  }

}


void Reader::set_current_header(const char *header)
{
  infoque[0].header = header;
}

void Reader::set_current_footer(const char *footer)
{
  infoque[0].footer = footer;
}


void Reader::set_textmode()
{
  infoque[0].need_head_signal = infoque[0].need_foot_signal = true;
}

int Reader::read_line(std::string &buf)
{
  static char rd_buf[MAX_LINE_LEN];

  if( infoque.size() == 0 ){
    return EOF;
  }

  fileinfo_t &info = infoque[0];

  if( info.header.size() > 0 ){
    buf = info.header;
    info.header = ""; // invalidate
    return buf.length();
  }
  if( info.need_head_signal ){
    info.need_head_signal = false; // incvalidate;
    buf = READER_INCLUDE_BEGIN + info.fnam ;
    return buf.length();
  }


  rd_buf[0] = '\0';
  
  int i=0;

  int c;
  while(1){
    c = get();
    if( c != EOF ){
      
      rd_buf[i++] = (unsigned char)c;
      if( c == '\n' ){
	info.linenum ++;
	break;
      }
    }else if( i>0 ){
      break;
    }else if(info.need_foot_signal ){
      info.need_foot_signal = false; //invalidate;
      buf = READER_INCLUDE_END;
      return strlen(READER_INCLUDE_END);
    }else if(info.footer.size() > 0 ){
      buf = info.footer;
      info.footer = ""; // invalidate
      return buf.length();
    }else if( infoque.size() > 0 ){
      infoque.pop_front();
      return read_line(buf);
    }else{
      return EOF;
      
    }
  
  }
  
  rd_buf[i]='\0';
  
  if( info.readopt & (READOPT_INCLUDE_HTML|READOPT_INCLUDE_TEXT|READOPT_TEXTMODE|READOPT_INCLUDE_NUMBERED) ){
    
    if( info.readopt==READOPT_TEXTMODE && strncmp(rd_buf,"@endtext",8)==0 ){
      info.readopt = 0; // ƥȥ⡼ɤȴ
      return read_line(buf);
    }
		       
    if( info.readopt & (READOPT_INCLUDE_TEXT|READOPT_INCLUDE_NUMBERED|READOPT_TEXTMODE) ){
      _text_chars_escape(rd_buf); // < > ʤɤ򥨥
    }
    
    if( info.readopt & READOPT_INCLUDE_NUMBERED ){
      buf = sourcemode_linenumber() + rd_buf;
    }
    else if( rd_buf[0] == '@' || rd_buf[0] == ';' ){
      char tmp[8];
      sprintf(tmp,"&#%d;",(int)rd_buf[0]);
      buf = tmp;
      buf += rd_buf+1;
    }
    else
      buf = rd_buf;
  }else
    buf = rd_buf;


  return i;
}


char* _parse_next(char *p,char **start_r)
{
  while( isspace(*p) )
    p++;

  char mode = 0;
  if( *p == '\"' || *p == '\''){   // " '
    mode = *p;
  }else if( *p == '(' ){           // (
    mode = ')';
  }

  if( mode )
    p++;

  *start_r  = p;
  
  if( mode )
    while( *p && *p!='\n' && *p!=mode )
      p++;
  else
    while( *p && ! isspace(*p) )
      p++;

  if( *p=='\0' )
    return NULL;
  
  //else
 
  *p = '\0';
  p++;
  while( isspace(*p) )
    p++;
  
  if( !*p || (*p==';' && *(p+1)==';') )
    return NULL; // ü,⤷,ĤƥȤǤ
  else
    return p;
 
  
}

int parse(const char *line,std::vector<std::string> &vec)
{
  char *b = new char[strlen(line)+1];
  strcpy(b,line);
  
  char *p = b;
  p++; // Ƭ@򥹥å
  
  vec.clear();
  char *next_p;
  do{
    next_p = _parse_next(p,&p);
    vec.push_back(p);
    //printf("'%s' ",p);
    p = next_p;
  }while( p != NULL );
  
  return vec.size();
}

static bool is_nullline(std::string &s)
{
  if( s.length() > 1 && s[0]==';' && s[1]==';' )
    return true;   // ȹԤʤΤǿ
  
  const char *p = s.c_str();
  while( *p && isspace(*p))
    p++;
  return *p=='\0'; // ⤷ϥڡǤʤԤǤΤǿ
}


#define TEMPFILENAME "_ezht_files_inc_.tmp"
static bool tempfile_created = false;

static void cmd_files(const std::vector<std::string> &args)
{
  if( args.size() < 4 ){
    error("@files: too few arguments");
    return;
  }

  std::string mode = args[1];
  if( mode != "text" &&
      mode != "numbered" ){
    error("@files: unknown mode `%s'",mode.c_str());
    return;
  }

  char curdir[1024];
  getcwd(curdir, sizeof curdir);
  std::string dir = args[2];

  // ǥ쥯ȥѹ
  if( 0 != chdir(dir.c_str()) ){
    error("@files: cannot enter directory `%s': %s",dir.c_str(),strerror(errno));
    return;
  }
  
  std::list<std::string> results;

  int n = (int) args.size();
  for(int i=3; i<n; i++){
    const char *pattern = args[i].c_str();
    if( 0 == ezht_glob(pattern,results) ){
      warn("@files: pattern `%s' not matched",pattern);
    }
  }

  chdir(curdir); // ᤹

  
  FILE *tmpf = fopen(TEMPFILENAME,"w");


  if( ! tmpf ){
    error("can't create temporary file: %s",strerror(errno));
    return;
  }
  
  tempfile_created = true;

  for(std::list<std::string>::iterator p=results.begin(); p!=results.end(); p++){
    std::string &fnam = *p;
    std::string path = dir + '/' + fnam;
    fprintf(tmpf,"@section \"%s\"\n",fnam.c_str());
    fprintf(tmpf,"@include \"%s\" %s\n",path.c_str(),mode.c_str());
    fprintf(tmpf,"@end\n");
  }
  
  fclose(tmpf);
  
  if( ! reader.open(TEMPFILENAME) ){
    error("@include: Unable to open temporary file `%s' ",TEMPFILENAME);
  }

}

void reader_tempfile_delete()
{
  if( tempfile_created )
    unlink(TEMPFILENAME);
}

const char * getline(bool accept_terminator)
{
  static std::string buf;
  
  while( EOF != reader.read_line(buf) ){
  
    if( buf.length() > 0 && buf[0]=='@' ){
      std::vector<std::string> vec;
      parse(buf.c_str(),vec);
      
      std::string cmd = vec[0];

      if( cmd == "include" )
	cmd_include(vec);
      else if( cmd == "jump" ){
	cmd_jump(vec,buf);
	return buf.c_str();
      }
      else if( cmd == "begintext" )
	cmd_begintext(vec);
      else if( cmd == "image" ){
	cmd_image(vec,buf);
	return buf.c_str();
      }else if( cmd == "refer" ){
	cmd_refer(vec,buf);
	return buf.c_str();
      }else if( cmd == "files" ){
	cmd_files(vec);
      }
      else if( cmd == "end" ){
	if( ! accept_terminator ){
	  error("misplaced @end");
	}else if( vec.size() != 1 ){
	  error("unknown parameter(s) in @end");
	}else
	  return "@end";
      }else{
	return buf.c_str();
      }
      
    }else if( is_nullline(buf) ){
      // ȹԤ⤷϶ԤʤΤ̵
      ;
    }else
      return buf.c_str();
  }

  
  
  return NULL;
}
