/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/


#ifndef _READER_H_
#define _READER_H_

#include <stdio.h>
#include <deque>
#include <string>
#include <vector>

class Reader;
extern Reader reader;

#define READOPT_INCLUDE_HTML     0x01
#define READOPT_INCLUDE_TEXT     0x02
#define READOPT_TEXTMODE         0x04
#define READOPT_INCLUDE_NUMBERED 0x08

#define READER_INCLUDE_BEGIN "\xff\xffIncludeBegin\xff\xff"
#define READER_INCLUDE_END   "\xff\xffIncludeEnd\xff\xff"

class fileinfo_t
{
 public:
  int  readopt;
  int  linenum;
  std::string fnam;
  std::string header;
  std::string footer;

  bool need_head_signal;
  bool need_foot_signal;

  std::deque<char> buf;
  std::deque<char>::iterator buf_pnt;
  
  fileinfo_t(std::string _fname, int _readopt){
    fnam = _fname;
    readopt = _readopt;
    linenum = 0;

    need_head_signal = need_foot_signal = false;
  }

  fileinfo_t(const fileinfo_t& src){
    fnam = src.fnam;
    readopt = src.readopt;
    linenum = src.linenum;

    need_head_signal = src.need_head_signal;
    need_foot_signal = src.need_foot_signal;
  }
  
  ~fileinfo_t(){
  }

};

class Reader
{
  enum{ MAX_LINE_LEN = 65536 };
  std::deque<fileinfo_t> infoque;
  
  int sourcemode_columns;
  int sourcemode_line_cnt;

  void set_sourcemode_columns();
  std::string sourcemode_linenumber();

  int get();
  void rewind();

public:
  
  enum{ Err=-2, End=-1};

  Reader();

  ~Reader();
  
  void set_current_readopt(int opt);

  bool open(const char *fnam,int opt=0,const char *filter=NULL);
  void set_current_header(const char *header);
  void set_current_footer(const char *footer);
  void set_textmode();

  int read_line(std::string &buf);

  const char* get_fnam();
  int get_linenum();

  std::string number_info();
};

const char * getline(bool accept_terminator=false);
int parse(const char *line,std::vector<std::string> &vec);

void reader_tempfile_delete();


#endif

