/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include "error.h"
#include "refer.h"

#include <map>

static std::map<std::string,std::string> references;

static void regist_reference_href(std::string name,std::string href)
{
  std::map<std::string,std::string>::iterator p = references.find(name);
  if( p != references.end() ){
    error("@refer: duplicate identifier `%s'",name.c_str()); 
    return;
  }
  
  references[name] = href;
}

void cmd_mark(std::vector<std::string> &args, VFile *out,std::string fname, std::string aname)
{
  if( ! out ){
    error("misplaced @mark");
    return;
  }
  if( args.size() > 3 ){
    error("@mark: too many arguments");
    return;
  }
  if( args.size() == 1 ){
    error("@mark: identifier required");
    return;
  }

  std::string ident = args[1];

  if( args.size() == 2 ){
    regist_reference_href(ident,fname+'#'+aname);
  }else{ // args.size() == 3
    if( args[2] != "here" ){
      error("@mark: unknown option `%s'",args[2].c_str());
      return;
    }
    static int id=0;
    char mk[16];
    sprintf(mk,"mark%d",id++);
    char tmp[64];
    sprintf(tmp,"<a name=\"%s\"></a>",mk);
    out->puts(tmp);
    regist_reference_href(ident,fname+'#'+mk);
  }
}

#define REFHEAD "\xff\xffrfhd-"

void cmd_refer(std::vector<std::string> &args, std::string &ret)
{
  if( args.size() < 3 ){
    error("@refer: too few arguments");
    return;
  }
  if( args.size() > 4 ){
    error("@refer: too many arguments");
    return;
  }

  std::string ident = args[1];
  std::string anchor= args[2];
  std::string attrs;
  if( args.size()==4 )
    attrs = args[3];
  
  char len_s[8]; sprintf(len_s,"%d:",ident.length());
  ret = (std::string)"<a href=\"" + REFHEAD + len_s + ident + "\" " + attrs + '>' + anchor + "</a>";
  
}


static bool find_reference_callback(VFile *vf, const char *line, std::string &modified_line_r, std::string*)
{
  modified_line_r = "";

  bool found=  false;
  const char *s;
  while( NULL != (s=strstr(line,REFHEAD)) ){
    found = true;
    int len = s - line;
    char *tmp = new char[len+1];
    memcpy(tmp,line,len);
    tmp[len] = '\0';
    modified_line_r += tmp;
    delete tmp;
    
    s += strlen(REFHEAD);
    int ident_len = atoi(s);
    while( isdigit(*s) ){
      s++;
    }
    s++;
    tmp = new char[ident_len+1];
    memcpy(tmp,s,ident_len);
    tmp[ident_len] = '\0';

    std::map<std::string,std::string>::iterator result = references.find(tmp);
    if( result == references.end() ){
      fprintf(stderr,"could not find the identifier `%s'\n",tmp);
      exit(-1);
    }
    delete tmp;
    
    modified_line_r += result->second;
    line = s + ident_len;
  }
  
  if( found ){
    modified_line_r += line;
    vf->chomp(); // ľιԤβԤȤ
  }

  return found;
}

void fixate_references()
{
  vf_each(find_reference_callback);
}

