/******************************************************************************
  ezht - Copyright (C) 2004 Tradcrafts
  This file is part of the ezht.
                                                                       
  ezht is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by the 
  Free Software Foundation;   either version 2 of the License,  or (at your 
  option) any later version.

  ezht is distributed in the hope that it will be useful, but 
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; see the file COPYING.  If not, write to the Free Software 
  Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
******************************************************************************/

#include "vfile.h"
#include "error.h"

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <deque>

VFile::VFile(const char *_fnam)
{
  is_clone = false;
  datas = new std::list<std::string>;
  cur = datas->end();
  fname = _fnam;
}

VFile::VFile(const VFile &src)
{
  is_clone = true;
  datas = src.datas;
  cur = src.cur;
  fname = src.fname;
}

VFile::~VFile()
{
  if( ! is_clone )
    delete datas;
}

void VFile::reset_filename(const char *new_fname)
{
  fname = new_fname;
}

VFile* VFile::clone()
{
  return 
    new VFile(*this);
}

const char* VFile::get_filename()
{
  return fname.c_str();
}

void VFile::chomp()
{

  if( cur == datas->begin() )
    return;

  std::list<std::string>::iterator p = cur;

  do{
    p--;
    std::string &s = *p;
    if( s[0] == '\t' ){
      if( s[s.length()-1] == '\n' ){
	char *tmp = new char[s.length()];
	memcpy(tmp,s.c_str(),s.length()-1);
	tmp[s.length()-1] = '\0';
	if( s.length() >= 2 && tmp[s.length()-2]=='\r' ){
	  // DOS,Win\r\nβԤξ
	  // \rͤФʤʤ (UNIXĶξ)
	  tmp[s.length()-2] = '\0';
	}
	s = tmp;
	delete tmp;
      }
      break;
    }
  }while( p != datas->begin() );

}

void VFile::puts(const char *s)
{
  char *buf = new char[strlen(s)+2];
  const char *p = s;

  while(1){
    while( *p && *p != '\n' ){
      p++;
    }

    buf[0] = '\t';

    if( *p == '\0' ){
      // Ԥ̵
      if( strlen(s) > 0 ){ 
	strcpy(buf+1,s);
	datas->insert(cur,buf);
      }
      break;
    }
    
    // *p == '\n'

    int len = p - s + 1;
    // Ԥ̵
    if( len > 0 && 0!=strcmp(s,"\n") && 0!=strcmp(s,"\r\n") ){
      memcpy(buf+1,s,len);
      buf[len+1] = '\0'; 
      datas->insert(cur, buf);
    }
    s = p = p+1;
      
  }

  delete buf;
}

void VFile::printf(const char *fmt, ...)
{
  static char buf[8192];
  va_list args;
  va_start(args,fmt);
  vsprintf(buf,fmt,args);
  va_end(args);

  VFile::puts(buf);
}

void VFile::add_index(const char *index_name)
{
  datas->insert(cur,index_name);
  datas->insert(cur,"");  // ɬ
}

bool VFile::seek(const std::string &index_name)
{
  std::list<std::string>::iterator p;
  for(p=datas->begin(); p!=datas->end(); p++)
    if( *p == index_name ){
      p++;
      cur = p;
      return true; // FOUND
    }

  return false; // NOT FOUND
}

static int compare(std::list<std::string> *datas,std::string fname)
{
  FILE *rd = fopen(fname.c_str(),"r");
  if( ! rd )
    return 1; // ե뤬¸ߤʤΤۤʤ

  std::deque<char> body;
  int c;
  while( EOF != (c=fgetc(rd)) )
    body.push_back((char)c);
  fclose(rd);
  

  std::deque<char>::iterator itr = body.begin();
  std::list<std::string>::iterator p;
  for(p=datas->begin(); p!=datas->end(); p++){
    std::string &str = *p;
    if( str[0] == '\t' ){
      const char *s = str.c_str()+1;
      while( *s ){
	if( itr == body.end() ||
	    *s++ != *itr++ ){
	  return 1; // ۤʤ
	}
      }
    }
  }
  
  return 0; //פ
}

void VFile::flush()
{
  if( 0==compare(datas,fname) ){
    fprintf(stderr,"[skipped] ");
    return;
  }

  FILE *fh = fopen(fname.c_str(),"w");
  if( ! fh ){
    error("can't create file `%s': %s",fname.c_str(), strerror(errno));
    return;
  }
  
  std::list<std::string>::iterator p;
  for(p=datas->begin(); p!=datas->end(); p++){
    std::string &str = *p;
    if( str[0] == '\t' ){
      if( EOF == fputs(str.c_str()+1,fh) ){
	fclose(fh);
	error("error occurred while writing in `%s': %s",fname.c_str(), strerror(errno));
	return;
      }
    }
  }

  fclose(fh);

  fprintf(stderr,"[wrote] ");
}

static void _put_body(const char *s,FILE *fh)
{
  fputc('-',fh);
  while( *s && *s!='\n' )
    fputc(*s++,fh);
  
  if( *s=='\n' ){
    fputc('\n',fh);
    if( *(s+1)=='\0' )
      return;
  }

  if( *s == '\0' )
    fputc('\n',fh);
  else
    _put_body(s+1,fh);
}


void VFile::flush2(FILE *out)
{
  fprintf(out,"+%s\n",fname.c_str());

  std::list<std::string>::iterator p;
  for(p=datas->begin(); p!=datas->end(); p++){
    std::string &str = *p;
    if( str[0] == '\t' ){
      _put_body(str.c_str()+1,out);
    }
  }
}

void VFile::rewind()
{
  cur = datas->begin();
}

void VFile::each( bool (*callback)(VFile*,const char*,std::string&,std::string*) )
{
  while( cur != datas->end() ){
    std::string &line = *cur;
    if( line[0] == '\t' ){
      std::string modified_line;
      if( (*callback)(this,line.c_str()+1,modified_line,&line) ){
	line = '\t' + modified_line;
      }
    }
    cur++;
  }

}

/////

static std::list<VFile*> vf_list;

VFile* vf_open(const char *fname)
{
  VFile *vf = new VFile(fname);
  vf_list.push_back(vf);
  return vf;
}

VFile* vf_search(const char *fname)
{
  std::list<VFile*>::iterator p;
  for(p=vf_list.begin(); p!=vf_list.end(); p++)
    if( 0 == strcmp(fname,(*p)->get_filename()) )
      return *p; // FOUND

  return NULL; // NOT FOUND
}


void vf_flushall()
{
  fprintf(stderr,"generating... "); fflush(stderr);

  std::list<VFile*>::iterator p;
  for(p=vf_list.begin(); p!=vf_list.end(); p++){
    fprintf(stderr,"%s",(*p)->get_filename()); fflush(stderr);
    (*p)->flush();
  }

  fprintf(stderr,"\ndone.\n");
}

void vf_flushall_to(FILE *out)
{
  std::list<VFile*>::iterator p;
  for(p=vf_list.begin(); p!=vf_list.end(); p++)
    (*p)->flush2(out);
}

///
void vf_each(bool (*callback)(VFile*,const char *,std::string &,std::string*))
{
  std::list<VFile*>::iterator p;
  for(p=vf_list.begin(); p!=vf_list.end(); p++){
    (*p)->rewind();
    (*p)->each(callback);
  }
}


